"""
Script that generates a dump of published MaveDB data.

Usage:
```
python3 -m mavedb.scripts.export_public_data
```

This generates a ZIP archive named `mavedb-dump.zip` in the working directory. the ZIP file has the following contents:
- main.json: A JSON file providing metadata for all of the published experiment sets, experiments, and score sets
- variants/
  - [URN].counts.csv (for each variant URN): The score set's variant count columns,
    sorted by variant number
  - [URN].scores.csv (for each variant URN): The score set's variant count columns,
    sorted by variant number

In the exported JSON metadata, the root object's `experimentSets` property gives an array of experiment sets.
Experiments are nested in their parent experiment sets, and score sets in their parent experiments.

The variant URNs used in filenames do not include the `urn:mavedb:` scheme identifier, so they look like
`00000001-a-1.counts.csv` and `00000001-a-1.scores.csv`, for instance.

Unpublished data and data sets licensed other than under the Create Commmons Zero license are not included in the dump,
and user details are limited to ORCID IDs and names of contributors to published data sets.
"""

import json
import logging
import os
from datetime import datetime, timezone
from itertools import chain
from typing import Callable, Iterable, TypeVar
from zipfile import ZipFile

from fastapi.encoders import jsonable_encoder
from sqlalchemy import select
from sqlalchemy.orm import lazyload, Session

from mavedb.lib.score_sets import get_score_set_counts_as_csv, get_score_set_scores_as_csv
from mavedb.models.experiment import Experiment
from mavedb.models.experiment_set import ExperimentSet
from mavedb.models.license import License
from mavedb.models.score_set import ScoreSet
from mavedb.view_models.experiment_set import ExperimentSetPublicDump

from mavedb.scripts.environment import script_environment, with_database_session

logger = logging.getLogger(__name__)

S = TypeVar("S")
T = TypeVar("T")


def filter_experiment_sets(experiment_sets: Iterable[ExperimentSet]) -> Iterable[ExperimentSet]:
    """
    Filter a list of experiment sets. Exclude any experiments with no score sets, then exclude experiment sets with no
    experiments.

    Filtering is done on the basis of the current contents of Experiment.score_set, which will have been loaded using a
    query that excludes unpublished score sets and those licensed other than under CC0.
    """
    return filter(filter_experiment_set, experiment_sets)


def filter_experiment_set(experiment_set: ExperimentSet):
    """
    Filter an experiment set. Exclude any experiments it contains that do not contain score sets, and return a value
    indicating whether any experiments remain.

    Filtering is done on the basis of the current contents of Experiment.score_set, which will have been loaded using a
    query that excludes unpublished score sets and those licensed other than under CC0.
    """
    experiment_set.experiments = list(filter_experiments(experiment_set.experiments))
    return len(experiment_set.experiments) > 0


def filter_experiments(experiments: Iterable[Experiment]) -> Iterable[Experiment]:
    """
    Filter a list of experiments, excluding any whose score_sets collection is empty.

    Filtering is done on the basis of the current contents of score_sets, which will have been loaded using a query that
    excludes unpublished score sets and those licensed other than under CC0.
    """
    return filter(lambda e: len(e.score_sets) > 0, experiments)


def flatmap(f: Callable[[S], Iterable[T]], items: Iterable[S]) -> Iterable[T]:
    return chain.from_iterable(map(f, items))


@script_environment.command()
@with_database_session
def export_public_data(db: Session):
    experiment_sets_query = db.scalars(
        select(ExperimentSet)
        .where(ExperimentSet.published_date.is_not(None))
        .options(
            lazyload(ExperimentSet.experiments.and_(Experiment.published_date.is_not(None))).options(
                lazyload(
                    Experiment.score_sets.and_(
                        ScoreSet.published_date.is_not(None), ScoreSet.license.has(License.short_name == "CC0")
                    )
                )
            )
        )
        .execution_options(populate_existing=True)
        .order_by(ExperimentSet.urn)
    )

    # Filter the stream of experiment sets to exclude experiments and experiment sets with no public, CC0-licensed score
    # sets.
    experiment_sets = list(filter_experiment_sets(experiment_sets_query.all()))

    # TODO To support very large data sets, we may want to use custom code for JSON-encoding an iterator.
    # Issue: https://github.com/VariantEffect/mavedb-api/issues/192
    # See, for instance, https://stackoverflow.com/questions/12670395/json-encoding-very-long-iterators.

    experiment_set_views = list(map(lambda es: ExperimentSetPublicDump.from_orm(es), experiment_sets))

    # Get a list of IDS of all the score sets included.
    score_set_ids = list(
        flatmap(lambda es: flatmap(lambda e: map(lambda ss: ss.id, e.score_sets), es.experiments), experiment_sets)
    )

    timestamp_format = "%Y%m%d%H%M%S"
    zip_file_name = f"mavedb-dump.{datetime.now().strftime(timestamp_format)}.zip"

    logger.info(f"Exporting public data set metadata to {zip_file_name}/main.json")
    json_data = {
        "title": "MaveDB public data",
        "asOf": datetime.now(timezone.utc).isoformat(),
        "experimentSets": experiment_set_views,
    }

    with ZipFile(zip_file_name, "w") as zipfile:
        # Write metadata for all data sets to a single JSON file.
        zipfile.writestr("main.json", json.dumps(jsonable_encoder(json_data)))

        # Copy the CC0 license.
        zipfile.write(os.path.join(os.path.dirname(__file__), "resources/CC0_license.txt"), "LICENSE.txt")

        # Write score and count files for each score set.
        num_score_sets = len(score_set_ids)
        for i, score_set_id in enumerate(score_set_ids):
            score_set = db.scalars(select(ScoreSet).where(ScoreSet.id == score_set_id)).one_or_none()
            if score_set is not None and score_set.urn is not None:
                logger.info(f"{i + 1}/{num_score_sets} Exporting variants for score set {score_set.urn}")
                csv_filename_base = score_set.urn.replace(":", "-")

                csv_str = get_score_set_scores_as_csv(db, score_set)
                zipfile.writestr(f"csv/{csv_filename_base}.scores.csv", csv_str)

                count_columns = score_set.dataset_columns["count_columns"] if score_set.dataset_columns else None
                if count_columns and len(count_columns) > 0:
                    csv_str = get_score_set_counts_as_csv(db, score_set)
                    zipfile.writestr(f"csv/{csv_filename_base}.counts.csv", csv_str)


if __name__ == "__main__":
    export_public_data()



import enum


class MappingState(enum.Enum):
    incomplete = "incomplete"
    processing = "processing"
    failed = "failed"
    complete = "complete"
    pending_variant_processing = "pending_variant_processing"
    not_attempted = "not_attempted"
    queued = "queued"



import logging
import os
from typing import Any

import httpx
from fastapi import APIRouter, Depends, HTTPException
from starlette.responses import JSONResponse

from mavedb.lib.authorization import require_current_user
from mavedb.lib.logging import LoggedRoute
from mavedb.lib.logging.context import logging_context, save_to_logging_context
from mavedb.lib.orcid import fetch_orcid_user
from mavedb.models.user import User
from mavedb.view_models import orcid

logger = logging.getLogger(__name__)

router = APIRouter(
    prefix="/api/v1/orcid",
    tags=["orcid"],
    responses={404: {"description": "Not found"}},
    route_class=LoggedRoute,
)

ORCID_CLIENT_ID = os.getenv("ORCID_CLIENT_ID")
ORCID_CLIENT_SECRET = os.getenv("ORCID_CLIENT_SECRET")


@router.get("/users/{orcid_id}", status_code=200, response_model=orcid.OrcidUser)
def lookup_orcid_user(
    orcid_id: str,
    user: User = Depends(require_current_user),
) -> Any:
    """
    Look an ORCID user up by ORCID ID.

    This capability is needed when adding contributors to an experiment or score set, who may not necessarily be MaveDB
    users.

    Access is limited to signed-in users to prevent abuse.
    """
    save_to_logging_context({"requested_resource": orcid_id})
    orcid_user = fetch_orcid_user(orcid_id)
    if orcid_user is None:
        return JSONResponse(
            status_code=404,
            content={},
        )
    else:
        return orcid_user


@router.post(
    "/token",
    status_code=200,
    response_model=orcid.OrcidAuthTokenResponse,
    responses={404: {}, 500: {}},
    include_in_schema=False,
)
async def get_token_from_code(*, request: orcid.OrcidAuthTokenRequest) -> Any:
    """
    Perform the second step of three-legged Oauth2. The user has signed into ORCID in the browser, and the client has
    obtained a code. We must now send this code to ORCID with our client secret, and ORCID will reply with a token.
    """
    async with httpx.AsyncClient() as client:
        url = "https://orcid.org/oauth/token"
        data = {
            "client_id": ORCID_CLIENT_ID,
            "client_secret": ORCID_CLIENT_SECRET,
            "code": request.code,
            "grant_type": "authorization_code",
            "redirect_uri": request.redirect_uri,
        }
        response = await client.post(url, data=data)
        if response.status_code == 200:
            data = response.json()
            token_type = data["token_type"]
            access_token = data["access_token"]
            expires_in = data["expires_in"]
            id_token = data["id_token"]

            save_to_logging_context({"token_type": token_type})

            if token_type is None or token_type.lower() != "bearer":
                logger.warning(
                    msg="Unexpected token type received from ORCID when exchanging code for token.",
                    extra=logging_context(),
                )

            return {
                "access_token": access_token,
                "expires_in": expires_in,
                "id_token": id_token,
                "token_type": token_type,
            }
        else:
            data = response.json()
            raise HTTPException(status_code=401, detail="Authentication error")



from sqlalchemy import Column, Enum, Integer

from mavedb.db.base import Base
from mavedb.models.enums.user_role import UserRole


class Role(Base):
    __tablename__ = "roles"

    id = Column(Integer, primary_key=True)
    name = Column(
        Enum(UserRole, create_constraint=True, length=32, native_enum=False, validate_strings=True),
        nullable=False,
    )



import logging
from typing import Optional

from sqlalchemy import func, or_
from sqlalchemy.orm import Session

from mavedb.lib.logging.context import logging_context, save_to_logging_context
from mavedb.models.contributor import Contributor
from mavedb.models.score_set import ScoreSet
from mavedb.models.target_gene import TargetGene
from mavedb.models.user import User
from mavedb.view_models.search import TextSearch

logger = logging.getLogger(__name__)


def search_target_genes(
    db: Session,
    owner_or_contributor: Optional[User],
    search: TextSearch,
    limit: Optional[int],
) -> list[TargetGene]:
    save_to_logging_context({"target_gene_search_criteria": search.dict()})

    query = db.query(TargetGene)

    if search.text and len(search.text.strip()) > 0:
        lower_search_text = search.text.strip().lower()
        query = query.filter(func.lower(TargetGene.name).contains(lower_search_text))
    if owner_or_contributor is not None:
        query = query.filter(
            TargetGene.score_set.has(
                or_(
                    ScoreSet.created_by_id == owner_or_contributor.id,
                    ScoreSet.contributors.any(
                        Contributor.orcid_id == owner_or_contributor.username
                    ),
                )
            )
        )

    query = query.order_by(TargetGene.name)
    if limit is not None:
        query = query.limit(limit)

    target_genes = query.all()
    if not target_genes:
        target_genes = []

    save_to_logging_context({"matching_resources": len(target_genes)})
    logger.debug(
        msg=f"Target gene search yielded {len(target_genes)} matching resources.",
        extra=logging_context(),
    )

    return target_genes



from datetime import date
from typing import TYPE_CHECKING, List, Optional

from sqlalchemy import Boolean, Column, Date, ForeignKey, Integer, String
from sqlalchemy.dialects.postgresql import JSONB
from sqlalchemy.event import listens_for
from sqlalchemy.ext.associationproxy import AssociationProxy, association_proxy
from sqlalchemy.orm import Mapped, relationship
from sqlalchemy.schema import Table

from mavedb.db.base import Base
from mavedb.lib.temp_urns import generate_temp_urn
from mavedb.models.contributor import Contributor
from mavedb.models.controlled_keyword import ControlledKeyword
from mavedb.models.doi_identifier import DoiIdentifier
from mavedb.models.experiment_controlled_keyword import ExperimentControlledKeywordAssociation
from mavedb.models.experiment_publication_identifier import ExperimentPublicationIdentifierAssociation
from mavedb.models.experiment_set import ExperimentSet
from mavedb.models.legacy_keyword import LegacyKeyword
from mavedb.models.publication_identifier import PublicationIdentifier
from mavedb.models.raw_read_identifier import RawReadIdentifier
from mavedb.models.user import User

if TYPE_CHECKING:
    from mavedb.models.score_set import ScoreSet

experiments_contributors_association_table = Table(
    "experiment_contributors",
    Base.metadata,
    Column("experiment_id", ForeignKey("experiments.id"), primary_key=True),
    Column("contributor_id", ForeignKey("contributors.id"), primary_key=True),
)

experiments_doi_identifiers_association_table = Table(
    "experiment_doi_identifiers",
    Base.metadata,
    Column("experiment_id", ForeignKey("experiments.id"), primary_key=True),
    Column("doi_identifier_id", ForeignKey("doi_identifiers.id"), primary_key=True),
)

experiments_legacy_keywords_association_table = Table(
    "experiment_keywords",
    Base.metadata,
    Column("experiment_id", ForeignKey("experiments.id"), primary_key=True),
    Column("keyword_id", ForeignKey("keywords.id"), primary_key=True),
)

# experiments_sra_identifiers_association_table = Table(
experiments_raw_read_identifiers_association_table = Table(
    "experiment_sra_identifiers",
    Base.metadata,
    Column("experiment_id", ForeignKey("experiments.id"), primary_key=True),
    Column("sra_identifier_id", ForeignKey("sra_identifiers.id"), primary_key=True),
)


class Experiment(Base):
    __tablename__ = "experiments"

    id = Column(Integer, primary_key=True)

    urn = Column(String(64), nullable=True, default=generate_temp_urn, unique=True, index=True)
    title = Column(String, nullable=False)
    short_description = Column(String, nullable=False)
    abstract_text = Column(String, nullable=False)
    method_text = Column(String, nullable=False)
    extra_metadata = Column(JSONB, nullable=False)

    private = Column(Boolean, nullable=False, default=True)
    approved = Column(Boolean, nullable=False, default=False)
    published_date = Column(Date, nullable=True)
    processing_state = Column(String, nullable=True)

    # TODO Remove this obsolete column.
    num_score_sets = Column("num_scoresets", Integer, nullable=False, default=0)
    score_sets: Mapped[List["ScoreSet"]] = relationship(back_populates="experiment", cascade="all, delete-orphan")

    experiment_set_id = Column(Integer, ForeignKey("experiment_sets.id"), index=True, nullable=True)
    experiment_set: Mapped[Optional[ExperimentSet]] = relationship(back_populates="experiments")

    created_by_id = Column(Integer, ForeignKey("users.id"), index=True, nullable=True)
    created_by: Mapped[User] = relationship("User", foreign_keys="Experiment.created_by_id")
    modified_by_id = Column(Integer, ForeignKey("users.id"), index=True, nullable=True)
    modified_by: Mapped[User] = relationship("User", foreign_keys="Experiment.modified_by_id")
    creation_date = Column(Date, nullable=False, default=date.today)
    modification_date = Column(Date, nullable=False, default=date.today, onupdate=date.today)
    contributors: Mapped[list["Contributor"]] = relationship(
        "Contributor", secondary=experiments_contributors_association_table, backref="experiments"
    )
    keyword_objs: Mapped[list["ExperimentControlledKeywordAssociation"]] = relationship(
        back_populates="experiment", cascade="all, delete-orphan"
    )
    legacy_keyword_objs: Mapped[list[LegacyKeyword]] = relationship(
        "LegacyKeyword", secondary=experiments_legacy_keywords_association_table, backref="experiments"
    )
    doi_identifiers: Mapped[list[DoiIdentifier]] = relationship(
        "DoiIdentifier", secondary=experiments_doi_identifiers_association_table, backref="experiments"
    )
    publication_identifier_associations: Mapped[list[ExperimentPublicationIdentifierAssociation]] = relationship(
        "ExperimentPublicationIdentifierAssociation", back_populates="experiment", cascade="all, delete-orphan"
    )
    publication_identifiers: AssociationProxy[List[PublicationIdentifier]] = association_proxy(
        "publication_identifier_associations",
        "publication",
        creator=lambda p: ExperimentPublicationIdentifierAssociation(publication=p, primary=p.primary),
    )

    # sra_identifiers = relationship('SraIdentifier', secondary=experiments_sra_identifiers_association_table, backref='experiments')
    raw_read_identifiers: Mapped[list[RawReadIdentifier]] = relationship(
        "RawReadIdentifier", secondary=experiments_raw_read_identifiers_association_table, backref="experiments"
    )

    # Unfortunately, we can't use association_proxy here, because in spite of what the documentation seems to imply, it
    # doesn't check for a pre-existing keyword with the same text.
    # keywords = association_proxy('keyword_objs', 'text', creator=lambda text: Keyword(text=text))

    # _updated_keywords: list[str] = None
    # _updated_doi_identifiers: list[str] = None

    @property
    # Original codes. Not sure this part.
    # def keywords(self) -> Dict[str, ControlledKeyword]:
    # Dict[str, ControlledKeyword] gets error Incompatible return value type in mypy
    def keywords(self) -> list[dict]:
        keyword_objs = self.keyword_objs or []
        keywords = []
        for keyword_assoc in keyword_objs:
            controlled_keyword = keyword_assoc.controlled_keyword
            keywords.append(
                {
                    "keyword": {
                        "key": controlled_keyword.key,
                        "value": controlled_keyword.value,
                        "vocabulary": controlled_keyword.vocabulary,
                        "special": controlled_keyword.special,
                        "description": controlled_keyword.description,
                    },
                    "description": keyword_assoc.description,
                }
            )
        return keywords

    @property
    def legacy_keywords(self) -> list[str]:
        legacy_keyword_objs = self.legacy_keyword_objs or []
        return [keyword_obj.text for keyword_obj in legacy_keyword_objs if keyword_obj.text is not None]

    async def set_legacy_keywords(self, db, keywords: Optional[list[str]]):
        if keywords:
            self.keyword_objs = [await self._find_or_create_legacy_keyword(db, text) for text in keywords]
        else:
            self.keyword_objs = []

    async def set_keywords(self, db, keywords: list):
        if keywords:
            self.keyword_objs = [
                ExperimentControlledKeywordAssociation(
                    experiment=self,
                    controlled_keyword=await self._find_keyword(
                        db, keyword_obj.keyword.key, keyword_obj.keyword.value, keyword_obj.keyword.vocabulary
                    ),
                    description=keyword_obj.description,
                )
                for keyword_obj in keywords
            ]
        else:
            self.keyword_objs = []

    # See https://gist.github.com/tachyondecay/e0fe90c074d6b6707d8f1b0b1dcc8e3a
    # @keywords.setter
    # async def set_keywords(self, db, keywords: list[str]):
    #     self._keyword_objs = [await self._find_or_create_keyword(text) for text in keywords]

    async def _find_or_create_legacy_keyword(self, db, keyword_text):
        keyword_obj = db.query(LegacyKeyword).filter(LegacyKeyword.text == keyword_text).one_or_none()
        if not keyword_obj:
            keyword_obj = LegacyKeyword(text=keyword_text)
        return keyword_obj

    async def _find_keyword(self, db, key: str, value: str, vocabulary: Optional[str]):
        query = (
            db.query(ControlledKeyword).filter(ControlledKeyword.key == key).filter(ControlledKeyword.value == value)
        )
        if vocabulary:
            query = query.filter(ControlledKeyword.vocabulary == vocabulary)
        controlled_keyword_obj = query.one_or_none()
        if controlled_keyword_obj is None:
            raise ValueError(f"Unknown keyword {key}:{value}")
        return controlled_keyword_obj


@listens_for(Experiment, "before_insert")
def create_parent_object(mapper, connect, target):
    if target.experiment_set_id is None:
        target.experiment_set = ExperimentSet(
            extra_metadata={},
            num_experiments=1,
            created_by=target.created_by,
            modified_by=target.modified_by,
            contributors=target.contributors,
        )



from itertools import chain
from typing import Any

import hgvs.dataproviders.uta
from cdot.hgvs.dataproviders import RESTDataProvider
from fastapi import APIRouter, Depends, HTTPException
from hgvs import parser, validator
from hgvs.exceptions import HGVSDataNotAvailableError, HGVSInvalidVariantError

from mavedb.deps import hgvs_data_provider

router = APIRouter(
    prefix="/api/v1/hgvs",
    tags=["hgvs"],
    responses={404: {"description": "Not found"}},
)


@router.get("/fetch/{accession}", status_code=200, response_model=str)
def hgvs_fetch(accession: str, hdp: RESTDataProvider = Depends(hgvs_data_provider)) -> str:
    """
    List stored sequences
    """
    try:
        return hdp.seqfetcher.fetch_seq(accession)
    except HGVSDataNotAvailableError as e:
        raise HTTPException(404, str(e))


@router.post("/validate", status_code=200, response_model=bool)
def hgvs_validate(variant: dict[str, str], hdp: RESTDataProvider = Depends(hgvs_data_provider)) -> bool:
    """
    Validate a provided variant
    """
    hp = parser.Parser()
    variant_hgvs = hp.parse(variant["variant"])

    try:
        valid = validator.Validator(hdp=hdp).validate(variant_hgvs, strict=False)
    except HGVSInvalidVariantError as e:
        raise HTTPException(400, str(e))
    else:
        return valid


@router.get("/assemblies", status_code=200, response_model=list[str])
def list_assemblies(hdp: RESTDataProvider = Depends(hgvs_data_provider)) -> list[str]:
    """
    List stored assemblies
    """
    return list(hdp.assembly_maps.keys())


@router.get("/{assembly}/accessions", status_code=200, response_model=list[str])
def list_accessions(assembly: str, hdp: RESTDataProvider = Depends(hgvs_data_provider)) -> list[str]:
    """
    List stored accessions
    """
    if assembly not in hdp.assembly_maps:
        raise HTTPException(404, f"Assembly '{assembly}' Not Found")

    return list(hdp.get_assembly_map(assembly_name=assembly).keys())


@router.get("/genes", status_code=200, response_model=list)
def list_genes():
    """
    List stored genes
    """
    # Even though it doesn't provide the most complete transcript pool, UTA does provide more direct
    # access to a complete list of genes which have transcript information available.
    return list(chain.from_iterable(hgvs.dataproviders.uta.connect()._fetchall("select hgnc from gene")))


@router.get("/genes/{gene}", status_code=200, response_model=dict[str, Any])
def gene_info(gene: str, hdp: RESTDataProvider = Depends(hgvs_data_provider)) -> dict[str, Any]:
    """
    List stored gene information for a specified gene
    """
    gene_info = hdp.get_gene_info(gene)

    if not gene_info:
        raise HTTPException(404, f"No gene info found for {gene}.")

    return gene_info


@router.get("/transcripts/gene/{gene}", status_code=200, response_model=list[str])
def list_transcripts_for_gene(gene: str, hdp: RESTDataProvider = Depends(hgvs_data_provider)) -> list[str]:
    """
    List transcripts associated with a particular gene
    """
    transcripts = set([tx_info["tx_ac"] for tx_info in hdp.get_tx_for_gene(gene)])

    if not transcripts:
        raise HTTPException(404, f"No associated transcripts found for {gene}.")

    return list(transcripts)


@router.get("/transcripts/{transcript}", status_code=200, response_model=dict[str, Any])
def transcript_info(transcript: str, hdp: RESTDataProvider = Depends(hgvs_data_provider)) -> dict[str, Any]:
    """
    List transcript information for a particular transcript
    """
    transcript_info = hdp.get_tx_identity_info(transcript)

    if not transcript_info:
        raise HTTPException(404, f"No transcript information found for {transcript}.")

    return transcript_info


@router.get("/transcripts/protein/{transcript}", status_code=200, response_model=str)
def convert_to_protein(transcript: str, hdp: RESTDataProvider = Depends(hgvs_data_provider)) -> str:
    """
    Convert a provided transcript from it's nucleotide accession identifier to its protein accession identifier
    """
    protein_transcript = hdp.get_pro_ac_for_tx_ac(transcript)

    if not protein_transcript:
        raise HTTPException(404, f"No protein transcript found for {transcript}.")

    return protein_transcript



from datetime import date
from typing import TYPE_CHECKING, List

from sqlalchemy import Column, Date, ForeignKey, Integer, String
from sqlalchemy.orm import Mapped, backref, relationship

from mavedb.db.base import Base

from .taxonomy import Taxonomy

if TYPE_CHECKING:
    from mavedb.models.target_gene import TargetGene


class TargetSequence(Base):
    __tablename__ = "target_sequences"

    id = Column(Integer, primary_key=True)
    sequence_type = Column(String, nullable=False)
    sequence = Column(String, nullable=False)
    label = Column(String, nullable=True)
    taxonomy_id = Column("taxonomy_id", Integer, ForeignKey("taxonomies.id"), index=True, nullable=True)
    taxonomy: Mapped[Taxonomy] = relationship(
        "Taxonomy",
        backref=backref("target_sequences", single_parent=True),
    )
    creation_date = Column(Date, nullable=False, default=date.today)
    modification_date = Column(Date, nullable=False, default=date.today, onupdate=date.today)

    target_genes: Mapped[List["TargetGene"]] = relationship(back_populates="target_sequence")



from typing import Any, List

from fastapi import APIRouter, Depends, HTTPException
from sqlalchemy import func, or_
from sqlalchemy.orm import Session

from mavedb import deps
from mavedb.lib.taxonomies import search_NCBI_taxonomy
from mavedb.models.taxonomy import Taxonomy
from mavedb.view_models import taxonomy
from mavedb.view_models.search import TextSearch

router = APIRouter(prefix="/api/v1/taxonomies", tags=["taxonomies"], responses={404: {"description": "Not found"}})


@router.get("/", status_code=200, response_model=List[taxonomy.Taxonomy], responses={404: {}})
def list_taxonomies(
    *,
    db: Session = Depends(deps.get_db),
) -> Any:
    """
    List taxonomies.
    """
    items = db.query(Taxonomy).order_by(Taxonomy.organism_name).all()
    return items


@router.get("/speciesNames", status_code=200, response_model=List[str], responses={404: {}})
def list_taxonomy_organism_names(
    *,
    db: Session = Depends(deps.get_db),
) -> Any:
    """
    List distinct species names, in alphabetical order.
    """

    items = db.query(Taxonomy).all()
    organism_names = map(lambda item: item.organism_name, items)
    return sorted(list(set(organism_names)))


@router.get("/commonNames", status_code=200, response_model=List[str], responses={404: {}})
def list_taxonomy_common_names(
    *,
    db: Session = Depends(deps.get_db),
) -> Any:
    """
    List distinct common names, in alphabetical order.
    """

    items = db.query(Taxonomy).all()
    common_names = map(lambda item: item.common_name, items)
    return sorted(list(set(common_names)))


@router.get("/{item_id}", status_code=200, response_model=taxonomy.Taxonomy, responses={404: {}})
def fetch_taxonomy(
    *,
    item_id: int,
    db: Session = Depends(deps.get_db),
) -> Any:
    """
    Fetch a single taxonomy by ID.
    """
    item = db.query(Taxonomy).filter(Taxonomy.id == item_id).first()
    if not item:
        raise HTTPException(status_code=404, detail=f"Taxonomy with ID {item_id} not found")
    return item


@router.get("/tax-id/{item_id}", status_code=200, response_model=taxonomy.Taxonomy, responses={404: {}})
def fetch_taxonomy_by_tax_id(
    *,
    item_id: int,
    db: Session = Depends(deps.get_db),
) -> Any:
    """
    Fetch a single taxonomy by tax_id.
    """
    item = db.query(Taxonomy).filter(Taxonomy.tax_id == item_id).first()
    if not item:
        raise HTTPException(status_code=404, detail=f"Taxonomy with tax_ID {item_id} not found")
    return item


@router.post("/search", status_code=200, response_model=List[taxonomy.Taxonomy])
async def search_taxonomies(search: TextSearch, db: Session = Depends(deps.get_db)) -> Any:
    """
    Search Taxonomy.
    If no search text, return the whole taxonomy list so that front end Taxonomy component can get data to show in dropdown button.
    """
    query = db.query(Taxonomy)

    if search.text and len(search.text.strip()) > 0:
        if search.text.isnumeric() is False:
            lower_search_text = search.text.strip().lower()
            query = query.filter(
                or_(
                    func.lower(Taxonomy.organism_name).contains(lower_search_text),
                    func.lower(Taxonomy.common_name).contains(lower_search_text),
                )
            )
        else:
            query = query.filter(Taxonomy.tax_id == int(search.text))
    items = query.order_by(Taxonomy.organism_name).all()

    if not items and search.text:
        search_taxonomy = await search_NCBI_taxonomy(db, search.text)
        if search_taxonomy:
            items = [search_taxonomy]
        else:
            items = []

    return items



NON_FIELD_ERRORS = "__all__"


class ValidationError(ValueError, AssertionError):
    def __init__(self, *args: object, triggers=None, custom_loc=None) -> None:
        super().__init__(*args)
        self.custom_loc = custom_loc

        self.triggering_exceptions = triggers



import logging
import logging.config
import os
import sys

from .config import load_stock_config
from .filters import canonical_only
from .formatters import MavedbJsonFormatter

WATCHTOWER_IMPORTED = False
try:
    from watchtower import CloudWatchLogHandler

    WATCHTOWER_IMPORTED = True
except ModuleNotFoundError:
    pass

LOG_CONFIG = os.environ.get("LOG_CONFIG")


def configure():
    """
    Configures logging for MaveDB.

    A stock configuration is loaded from the ``mavedb/logging/configurations/*.yaml``
    files, chosen based on the value of ``LOG_CONFIG``.

    Python library warnings are captured and logged at the ``WARNING`` level.
    Uncaught exceptions are logged at the ``CRITICAL`` level before they cause
    the process to exit.
    """
    stock_config = load_stock_config(LOG_CONFIG if LOG_CONFIG else "default")
    logging.config.dictConfig(stock_config)

    # Log library API warnings.
    logging.captureWarnings(True)

    # Log any uncaught exceptions which are about to cause process exit.
    sys.excepthook = lambda *args: logging.getLogger().critical("Uncaught exception:", exc_info=args)  # type: ignore

    # Formatter and handler are un-configurable via file config.
    cw_is_enabled = False
    root_logger = logging.getLogger("root")

    if WATCHTOWER_IMPORTED:
        for handler in root_logger.handlers:
            if isinstance(handler, CloudWatchLogHandler):
                handler.addFilter(canonical_only)
                handler.formatter = MavedbJsonFormatter()

                cw_is_enabled = True

    if not cw_is_enabled:
        root_logger.info("CloudWatch log handler is not enabled. Canonical logs will only be emitted to stdout.")
